/* ---------- SUFF_0_3a: Final Version. -----------    
  
   This simplistic UCI-chess engine was my very first attempt in coding any 
   useful program.
   The topic 'chess' came into my mind, because I'm a passionate (but
   weak) chess player (~1850 Elo Rating).
   Suff_0_3a is - most likely - a good example for how a chess engine (and any
   other program in general!) SHOULD NOT BE CODED.
   Its source code is a demonstration of every known (and perhaps - until today - even unknown)
   programming anti-pattern, confusing spaghetti-loops, almost no useful comments, 
   plenty of global data, misnamed functions, magic numbers ...

   - That's why I'm somewhat ashamed of my inferior programming style. -

   Nevertheless I've decided to make Suff_0_3a (.exe as well as .src) a 
   PUBLIC DOMAIN chess engine, both the .exe and the source files.
   At least MS-VC++ 2008 Express compiled this chaotic 
   source code without errors, and Suff_0_3a was then usable inside Fritz_11's 
   and Arena's GUI to play chess against it.
   (With WinXP or Vista one will sometimes need to install a special
	c-runtime.dll from M$'s page to run Suff_0_3a.)
   It seems to me that I have successfully eliminated most runtime-bugs, but I'm
   not an expert programmer, that's why there could still be some minor bugs. -

   I didn't take much time for planning the program's design, I just sat
   down and coded whatever came to my mind and what I thought could be suitable
   to write a chess engine.
   As an amateur programmer it was a surprise - even to me -
   that after ~6 weeks of working on Suff_0_3a, the source compiled without errors
   and Suff_0_3a could be loaded to Fritz_11's GUI (to play chess against Suff).
   Some compilers won't compile C++ code with .h headers, these #include-lines-of-code 
   can simply be replaced by the 'correct' <cstring> + namespace code.
   The source code is confusing enough to disencourage modifications however.
   
   A short 'technical' overview:
   
   The [struct situation] is used to hold a position's state in memory.
   - It contains a char[8][8] array, which stores informations about whether
     a square contains a piece, its type, and its 'owner' ('color').
   - The iSquare[16] whitePieces and iSquare[16] blackPieces store the position
	 of every piece for the both players. The index [0-15] of these members
	 is used as an 'ID' of the piece (When generating a move, these 'IDs' are
	 stored in the move-struct's active_id and captured_id members.
	 The execute_move() needs these IDs to remove the captured piece, 
	 and this ID is also necessary for en passant capture moves,
	 where the destination square and the square of the 'captured piece' are not
	 the same.
   - The extraInfo member stores, whether White or Black may move + castling rights.
   - epSquare = en-passant Square
   - wKing, bKing = are needed very often, that's why I stored their IDs as extra members
	 of the situation-struct
     (otherwise I'd need to iterate the iSquare[16] arrays too often).

   Short description of important functions:
   find_moves() will generate every possible move for a position, and store them
	 to its move_table parameter.
   find_quickly() will only try to find at least one legal move and is used by the
     evaluate() function to determine whether a position at REAL_DEPTH is mate
	 or stalemate.
   execute_move() simply executes a move on its situation* parameter.
   quick_check() will analyze whether a square is controled by an opponent's piece.
   is_under_check() similar quick_check(), but addionally returns more accurate 
					informations about those pieces, which control the square.
   THINK() = the root of the game tree.
   buck_sort(), hyper_sort() ... they pre-sort the move_tables returned by find_moves,
							 for pruning purposes
   Second_ALPHA_BETA, Third_ALPHA_BETA ... () are special Versions of the AlphaBeta
							 algorithm for each search depth. They call different
							 pre-sorting functions (at low search depths I can afford
							 to call more accurate, time-consuming pre-sort functions.
							 At higher search depths I need to call faster pre-sort
							 functions (which, however, are less accurate than f.e.
							 hyper_sort))
   search_best_move() is called in response to a 'position' command and prepares
					  the search
   init() creates the most important resources + prerequisites of Suff in memory
   evaluate_opening(), evaluate() there are two different functions for evaluation,
								  this depends on the total number of pieces on the
								  board, which of the two is called at REAL_DEPTH
   the main() Function itself contains the code which communicates with the GUI
   create_internal_board() will initialize the Global situation-struct 'CurrentBoard' with
			  a FenString   
   there are also a few mini-functions and define-macros, which I needed at many occassions

   Most data is stored in Globals: there are MAX_DEPTH+2 situation-structs, MAX_DEPTH+2
   move_tables, MAX_DEPTH+2 sorting containers for hyper_sort, buck_sort, mediocre_sort etc.
   such that every AlphaBeta function at every search depth has its 'own' and 'independant'
   data of these types for manipulation (e.g. executing moves etc.)

   Algorithms:
   - Basic Alpha_Beta Pruning with pre-sorted move_tables
   - instead of a 'real' Quiescent Search I used a
	 'horizon penalty' which will only consider re-capture possiblities by an opponent, if
     the 'last move' before reaching REAL_DEPTH was a capturing move.
   - moves are pre-sorted by an algorithm similar (but not the same) to BucketSort
   - a tree-like data structure stores some variations that have been evaluated earlier,
     this could occassionally help to avoid some child nodes due to transpositions.
   
   - ? All in all Suff is not a sophisticated chess engine at all, as an amateur programmer
     I wasn't familiar with professional algorithms like Zobrist's hash function, Principal
	 Variation Search, __int64 Bitboards ... at the time when I coded Suff.
	 However, I've learned a lot about chess programming (and programming in general) due 
	 to Suff and its programming anti-patterns I made here (especially Suff's inferior 
	 modularization), which made it hard and finally impossible to improve the program.
	 For my next chess engine (I'm currently working on) I'll try to avoid these stupid 
	 programming anti-patterns, so maybe it will play stronger moves and be easier to
	 improve (and debug).

				Clemens Pruell, c_pruell@gmx.at , July 28, 2009, c_pruell@gmx.at

	-------	*/

#ifndef MYCHESS_CONSTANTS
#define MYCHESS_CONSTANTS

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <math.h>

#define TRUE 1
#define FALSE 0

#define NUMBER_OF_FILES 8
#define NUMBER_OF_RANKS 8

#define NUMBER_OF_DIRECTIONS 9
#define INVALID_SQUARE 127
#define STANDARD_DIR 0
#define SOUTHWEST 1
#define WEST 2
#define NORTHWEST 3
#define NORTH 4
#define NORTHEAST 5
#define EAST 6
#define SOUTHEAST 7
#define SOUTH 8

#define KING_ID 1
#define QUEEN_ID 2
#define ROOK_ID 3
#define BISHOP_ID 4
#define KNIGHT_ID 5
#define PAWN_ID 6
#define PIECE_MASK 7
#define NUMBER_OF_DIFFERENT_PIECES 6
#define MAXIMUM_NUMBER_OF_PIECES 16
#define GET_PIECE_TYPE(x) ((x) & PIECE_MASK)
#define WHITE_PAWN (PAWN_ID | WHITE)
#define BLACK_PAWN (PAWN_ID | BLACK)
#define WHITE_ROOK (ROOK_ID | WHITE)
#define BLACK_ROOK (ROOK_ID | BLACK)

#define WHITE 16
#define BLACK 32
#define COLOR_MASK 48
#define GET_PIECE_COLOR(x) ((x) & COLOR_MASK)

#define WKCASTLE 1
#define WQCASTLE 2
#define BKCASTLE 4
#define BQCASTLE 8
#define ANY_WCASTLE 3
#define ANY_BCASTLE 12
#define INITIAL_WHITE_KING_FILE 4
#define INITIAL_WHITE_KING_RANK 0
#define INITIAL_BLACK_KING_FILE 4
#define INITIAL_BLACK_KING_RANK 7

#define NO_CAPTURE 16
#define MOVE_TABLE_DIMENSION 100


/* Structs */
typedef char iSquare;

struct situation{
	char square[NUMBER_OF_FILES][NUMBER_OF_RANKS];
	char extraInfo; /* stores side to move and castling availabilties */
	iSquare epSquare;
	iSquare white_Pieces[MAXIMUM_NUMBER_OF_PIECES];
	iSquare black_Pieces[MAXIMUM_NUMBER_OF_PIECES];

	char king_w, king_b;
	/* very important to check for allowed moves */
	char hMoves;
};

struct Destination{
	char number_of;
	iSquare* dest_Square;
};

struct chess_vector{
	int vec_x;
	int vec_y;
};


struct move{
	iSquare start;
	iSquare dest;
	char active_id;
	char captured_id;
	char promote;
	short quality;
};

struct move_table{
	unsigned short number_of;
	char mate;
	char check;
	move stored_move[MOVE_TABLE_DIMENSION];
};



/* (1) Allocates Memory for situation* CurrentBoard */
int init();

/* setup of the internal Board based on the information given by a FENSTRING */
int create_internal_board(char* fenString);

/* converts a square that is given by algebraic notation into internal repres. */
iSquare alg_convert(char aFile, char aRank);

/* converts a square that is given by two indizes into internal repres. */
iSquare index_convert(int iFile, int iRank);

/* converts an iSquare to algebraic characters */
void convert_to_alg(iSquare src, char* destFile, char* destRank);

/* the buffer must be at least 3 bytes of size */
void convert_to_string(iSquare src, char* buffer);

inline char GET_FILE(iSquare src);

inline char GET_RANK(iSquare src);

/* initializes the fixed_Destination array. */
int create_Destinations(int piece_type, int file, int rank);

iSquare add_vector(iSquare start_point, int vec_x, int vec_y);

chess_vector subtract(iSquare top, iSquare base);

void multiply_vector(chess_vector* source, chess_vector* dest, int scalar);

int is_Corner_Square(int file, int rank);

int en_passant_extra(situation* tested, iSquare active);


#define NUMBER_OF_CRITERIA 7
#define MEDIOCRE_CRITERIA 19
#define MEDIOCRE_MOVE 18
#define HYPER_CRITERIA 26

#define MAX_DEPTH 7
#define BUCKET_DEPTH 4
#define MEDIOCRE_DEPTH 5

typedef move_table bucket_criteria[MEDIOCRE_CRITERIA];
typedef move_table  accurate_criteria[HYPER_CRITERIA];

int init_move_tables();

void buck_sort(short table_index);
void hyper_sort(short table_index);
void first_sort(short table_index);


#define CHECK_COUNTER_MASK 15L
#define CHECK_SOURCE_MASK 255L
#define CHECK_DIR_MASK 240
#define GET_CHECK_DIRECTION(x) (((x) & CHECK_DIR_MASK) >> 4)

/* these may only be called after the necessary initializing is done */
long is_under_check(situation* examined, iSquare xy);

/* the lock_indicator must be an array of 9 shorts */
int lock_directions(situation* examined, iSquare piece_loc,
					 short* lock_indicator);


int find_moves(situation* examined, move_table* results);

int find_quickly(situation* examined, move_table* results);

int quick_check(situation* examined, iSquare xy);

void move_to_alg(move* src, char* destString);

void alg_to_move(char* src, move* destMove);

int get_Move_index(move* question, move_table* setOfMoves);

char execute_move(situation* board, move* xy);


/* CONSTANTS, STRUCTS and functions for evaluation */
#define PAWN_VALUE 100
#define KNIGHT_VALUE 300
#define BISHOP_VALUE 300
#define LIGHT_PIECE_VALUE 300
#define ROOK_VALUE 450
#define QUEEN_VALUE 900
#define KING_VALUE 15000
#define STALE_MATE 0

#define NO_CASTLE 15
#define CASTLING_LOST 43
#define CENTER_PAWN 26
#define HALF_CENTER 11

#define SAFE_KING 8
#define OPEN_FILE 13
#define DANGEROUS_FILE 18
#define OPEN_RANK 7
#define DANGEROUS_RANK 15
#define DEVELOPPED_PIECE 13
#define GOOD_KNIGHT 11
#define BAD_KNIGHT 8

#define SIDE_TO_MOVE_BONUS 11
#define BISHOP_PAIR_OPENING 13
#define BISHOP_PAIR_MIDGAME 21
#define BISHOP_PAIR_ENDING 75
#define DANG_PAWN 17
#define PASSED_PAWN 27
#define PASSED_PAWN_BONUS 2
#define DANG_BISHOP 11
#define MULTIPLE_PAWNS 3
#define ISOLATED_PAWN 1
#define QUEEN_TOO_SOON 3
#define ROOK_MATE_FACTOR 18
#define ROOK_MATE_FACTOR_2 18
#define QUEEN_MATE_FACTOR 21
#define QUEEN_MATE_FACTOR_2 18
#define OPENING 24
#define ENDING 12
#define DANGEROUS_QUEEN 5
#define ACTIVE_KING_ENDGAME 50
#define UNDEVELOPPED_PIECE 28
#define CENTER 20
#define ROOK_FILE_BONUS 3
#define DEVELOPPED_BISHOP 15

#define MISTAKE_WINDOW 15
#define RANDOM_WINDOW 7
#define NUMBER_OF_REFERENCES 6

struct ev_Table{
	char pawn_count_w[NUMBER_OF_FILES];
	char pawn_mask_w[NUMBER_OF_FILES];
	char pawn_most_advanced_w[NUMBER_OF_FILES];
	char pawn_count_w_rank_id[NUMBER_OF_FILES];

	char pawn_count_b[NUMBER_OF_FILES];
	char pawn_mask_b[NUMBER_OF_FILES];
	char pawn_most_advanced_b[NUMBER_OF_FILES];
	char pawn_count_b_rank_id[NUMBER_OF_FILES];

	char pawn_count_w_r[NUMBER_OF_RANKS];
	char pawn_count_b_r[NUMBER_OF_RANKS];

	char rook_count_b[NUMBER_OF_FILES];
	char rook_count_w[NUMBER_OF_FILES];
	char rook_count_b_r[NUMBER_OF_RANKS];
	char rook_count_w_r[NUMBER_OF_RANKS];	
	
	char piece_count_w, piece_count_b;
	iSquare w_queen, b_queen;
	iSquare w_king, b_king;
	iSquare w_bishop[2];
	iSquare b_bishop[2];
	char w_bishop_count, b_bishop_count;
	char w_queen_count, b_queen_count;
	char w_rook, b_rook;
	char w_rook_count, b_rook_count;
};


struct ev_Table_Opening{	
	char pawn_mask_w[NUMBER_OF_FILES];
	char pawn_mask_b[NUMBER_OF_FILES];
	iSquare w_queen, b_queen;
	iSquare w_king, b_king;
	char w_bishop_count, b_bishop_count;
	char w_queen_count, b_queen_count;
};

int evaluate_moves(int your_depth, situation* pos_searched);

short THINK(situation* pos, int your_depth);

short Second_ALPHA_BETA(situation* pos, int your_depth, short alpha, short beta);

short Third_ALPHA_BETA(situation* pos, int your_depth, short alpha, short beta);

short Bucket_ALPHA_BETA(situation* pos, int your_depth, short alpha, short beta);

short Mediocre_ALPHA_BETA(situation* pos, int your_depth, short alpha, short beta);

short Standard_ALPHA_BETA(situation* pos, int your_depth, short alpha, short beta);

int search_best_move(char* result);

#endif

/*	to analyze, which functions are still too slow:
	QueryPerformanceCounter()
#include <windows.h> */
